'use strict';


const {Gio, GObject, GLib, St} = imports.gi;
const Main = imports.ui.main;
const MainPanel = imports.ui.main.panel;
const PopupMenu = imports.ui.popupMenu;
const PanelMenu = imports.ui.panelMenu;
const MessageTray = imports.ui.messageTray;
const QuickSettings = imports.ui.quickSettings;
const QuickSettingsMenu = imports.ui.main.panel.statusArea.quickSettings;
const Clutter = imports.gi.Clutter;
const ExtensionUtils = imports.misc.extensionUtils;

const Me = ExtensionUtils.getCurrentExtension();
const Settings = ExtensionUtils.getSettings();

const MAIN_ICON = 'fcitx-fullwidth-active';

const urgencyMapping = {
    0: MessageTray.Urgency.LOW,
    1: MessageTray.Urgency.NORMAL,
    2: MessageTray.Urgency.CRITICAL,
};


const FeatureMenuToggle = GObject.registerClass(
class FeatureMenuToggle extends QuickSettings.QuickMenuToggle {
    _init() {
        super._init({
            label: Me.metadata.name,
            iconName: MAIN_ICON,
            toggleMode: true,
        });

        Settings.bind('timer-enabled',
            this, 'checked',
            Gio.SettingsBindFlags.DEFAULT);

        this.menu.setHeader(MAIN_ICON, Me.metadata.name,
            Me.metadata.description);

        this.menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());

        const settingsItem = this.menu.addAction('Settings',
            () => ExtensionUtils.openPrefs());

        settingsItem.visible = Main.sessionMode.allowSettings;
        this.menu._settingsActions[Me.uuid] = settingsItem;
    }
});


const FeatureIndicator = GObject.registerClass(
class FeatureIndicator extends QuickSettings.SystemIndicator {
    _init() {
        super._init();

        this._indicator = this._addIndicator();
        this._indicator.icon_name = MAIN_ICON;

        Settings.bind('timer-enabled',
            this._indicator, 'visible',
            Gio.SettingsBindFlags.DEFAULT);

        this.quickSettingsItems.push(new FeatureMenuToggle());

        this.connect('destroy', () => {
            this.quickSettingsItems.forEach(item => item.destroy());
        });

        QuickSettingsMenu._indicators.add_child(this);
        QuickSettingsMenu._addItems(this.quickSettingsItems);
    }
});


class Extension {
    constructor() {
        this._indicator = null;
        this._timer = null;
        this._timerLabel = null;
        this._notifyLoopId = null;
        this._timerSec = null;
        this._startSec = null;
    }

    _notify() {
        let messageText = Settings.get_string('message-text') || 'It is time to stretch your back!';
        let iconName = Settings.get_string('icon-name') || 'dialog-messages';
        let urgencyLevel = Settings.get_int('urgency-level');
        let mappedUrgency = urgencyMapping[urgencyLevel];

        let source = new MessageTray.Source(Me.metadata.name, iconName);
        Main.messageTray.add(source);
        let notification = new MessageTray.Notification(source, Me.metadata.name, messageText);
        notification.setUrgency(mappedUrgency);
        notification.setTransient(true);
        source.showNotification(notification);

        let playSound = Settings.get_boolean('play-sound');
        if (playSound) {
            let soundName = Settings.get_string('sound-name') || 'complete';
            let player = global.display.get_sound_player();
            player.play_from_theme(soundName, Me.metadata.name, null);
        }
    }


    _formatTime(totalSeconds) {
        let totalMinutes = Math.floor(totalSeconds / 60);
        let seconds = totalSeconds % 60;
        let hours = Math.floor(totalMinutes / 60);
        let minutes = totalMinutes % 60;

        if (hours < 10) {
            hours = "0" + hours;
        }

        if (minutes < 10) {
            minutes = "0" + minutes;
        }

        if (seconds < 10) {
            seconds = "0" + seconds;
        }

        if (hours > 0) {
            return `${hours}:${minutes}:${seconds}`;
        } else {
            return `${minutes}:${seconds}`;
        }
    }


    _addTimer() {
        this._timer = new PanelMenu.Button(0.0, Me.metadata.name, false);
        this._timerLabel = new St.Label({y_expand: true, y_align: Clutter.ActorAlign.CENTER });
        this._timer.add_child(this._timerLabel);
        Main.panel.addToStatusArea(Me.metadata.name, this._timer);
    }

    _removeTimer() {
        this._timer.destroy();
        this._timer = null;
    }


    _startTimer() {
        this._startSec = Math.floor(new Date().getTime() / 1000);
        this._timerSec = Settings.get_int('interval-min') * 60;

        let showTimer = Settings.get_boolean('show-timer');
        if (showTimer) {
            this._addTimer();
        }

        this._notifyLoopId = GLib.timeout_add_seconds(GLib.PRIORITY_DEFAULT, 1, () => {
            this._updateTimer();
            return GLib.SOURCE_CONTINUE;
        });
    }


    _updateTimer() {
        this._lastSec = Math.floor(new Date().getTime() / 1000);
        let diffSec = this._lastSec - this._startSec;

        let showTimer = Settings.get_boolean('show-timer');
        if (showTimer) {
            this._timerLabel.set_text(this._formatTime(this._timerSec - diffSec));
        }

        if (diffSec >= this._timerSec) {
            this._notify();
            this._startSec = Math.floor(new Date().getTime() / 1000);
            if (showTimer) {
                this._timerLabel.set_text(this._formatTime(this._timerSec));
            }
        }
    }


    _stopTimer() {
        GLib.Source.remove(this._notifyLoopId);
        this._notifyLoopId = null;
        if (this._timer) {
            this._removeTimer();
        }
        this._timerSec = null;
        this._startSec = null;
    }


    enable() {
        this._indicator = new FeatureIndicator();

        Settings.connect('changed::timer-enabled', function(Settings, key) {
            if (Settings.get_boolean(key)) {
                this._startTimer();
            } else {
                this._stopTimer();
            }
        }.bind(this));

        Settings.connect('changed::show-timer', function(Settings, key) {
            if (Settings.get_boolean('timer-enabled')) {
                if (Settings.get_boolean(key)) {
                    this._addTimer();
                } else {
                    this._removeTimer();
                }
            }
        }.bind(this));
    }


    disable() {
        /*
        unlock-dialog is required to keep the notification
        loop running even when the user locked the screen

        and if the user just forgot to disable the timer
        they should be able to do so from the locked screeen
        */
        if (this._notifyLoopId) {
            GLib.Source.remove(this._notifyLoopId);
            this._notifyLoopId = null;
        }

        Settings.set_boolean('timer-enabled', false);
        Settings.disconnect('changed');

        if (this._timer) {
            this._removeTimer();
        }

        this._indicator.destroy();
        this._indicator = null;
    }
}


function init() {
    return new Extension();
}
